import numpy

import chainer
from chainer.functions.activation import sigmoid
from chainer.functions.activation import tanh
from chainer.functions import split_axis
from chainer import link
from chainer.links.connection import linear

from chainer import cuda
from chainer import function
from chainer.utils import type_check

if cuda.cudnn_enabled:
    cudnn = cuda.cudnn
    libcudnn = cudnn.cudnn
    _mode = libcudnn.CUDNN_ACTIVATION_SIGMOID


class SigmoidAPLusBByH(function.Function):

    """SigmoidAPLusBByH function."""

#     def __init__(self):
#         self.use_cudnn = use_cudnn

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 3)
        type_check.expect(in_types[0].dtype == numpy.float32)
        type_check.expect(in_types[1].dtype == numpy.float32)
        type_check.expect(in_types[2].dtype == numpy.float32)

    def forward_cpu(self, x):
        self.sigma_a_plus_b = (numpy.tanh((x[0] + x[1]) * 0.5) * 0.5 + 0.5)
        return x[2] * self.sigma_a_plus_b,

    def forward_gpu(self, x):
        self.sigma_a_plus_b, y = cuda.elementwise(
            'T x1, T x2, T x3', 'T sigma_a_plus_b, T y',
            '''
                sigma_a_plus_b = tanh((x1 + x2) * 0.5) * 0.5 + 0.5;// 1 / (1 + exp(-(x1 + x2)));
                y = x3 * sigma_a_plus_b;
                ''',
            'sigmoid_a_plus_b_by_h_fwd')(x[0], x[1], x[2])
        return y,

    def backward_cpu(self, x, gy):
        gy_by_sigma_a_plus_b = gy[0] * self.sigma_a_plus_b
        deriv1 = x[2] * gy_by_sigma_a_plus_b * (1 - self.sigma_a_plus_b)
        return deriv1, deriv1, gy_by_sigma_a_plus_b

    def backward_gpu(self, x, gy):
        gx1, gx2, gx3 = cuda.elementwise(
            'T sigma_a_plus_b, T h, T gy', 'T gx1, T gx2, T gx3',
            '''
            gx3 = gy * sigma_a_plus_b;
            gx1 = h * gx3 * (1-sigma_a_plus_b);
            gx2 = gx1;
            ''',
            'sigmoid_bwd')(self.sigma_a_plus_b, x[2], gy[0])
        return gx1, gx2, gx3,


def sigm_a_plus_b_by_h_fast(x1, x2, x3):
    """ sigm_a_plus_b_by_h_fast
    """
    return SigmoidAPLusBByH()(x1, x2, x3)


class ComputeOutputGRU(function.Function):

    """SigmoidAPLusBByH function."""

#     def __init__(self):
#         self.use_cudnn = use_cudnn

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 5)
        type_check.expect(in_types[0].dtype == numpy.float32)
        type_check.expect(in_types[1].dtype == numpy.float32)
        type_check.expect(in_types[2].dtype == numpy.float32)
        type_check.expect(in_types[3].dtype == numpy.float32)
        type_check.expect(in_types[4].dtype == numpy.float32)

    def forward_cpu(self, x):
        z_x, z_h, h_x, h, hh = x
        z = 1.0 / (1 + numpy.exp(- (z_x + z_h)))
        h_bar = numpy.tanh(h_x + hh)
        h_new = (1 - z) * h + z * h_bar

        self.z = z
        self.h_bar = h_bar
        return h_new,

    def forward_gpu(self, x):
        z_x, z_h, h_x, h, hh = x
        self.z, self.h_bar, h_new = cuda.elementwise(
            'T z_x, T z_h, T h_x, T h, T hh',
            'T z, T h_bar, T h_new',
            '''
                z = tanh((z_x + z_h) * 0.5) * 0.5 + 0.5;
                //z = 1.0/ ( 1 + exp(- (z_x + z_h)));
                h_bar = tanh(h_x + hh);
                h_new = (1 - z) * h + z * h_bar;
                ''',
            'compute_output_gru_fwd')(z_x, z_h, h_x, h, hh)
        return h_new,

    def backward_cpu(self, x, gy):
        z_x, z_h, h_x, h, hh = x
        g_h = (1 - self.z) * gy[0]
        g_hh = self.z * (1 - self.h_bar * self.h_bar) * gy[0]
        g_h_x = g_hh
        g_z_x = g_h * self.z * (self.h_bar - h)
        g_z_h = g_z_x
        return g_z_x, g_z_h, g_h_x, g_h, g_hh

    def backward_gpu(self, x, gy):
        z_x, z_h, h_x, h, hh = x
        g_z_x, g_z_h, g_h_x, g_h, g_hh = cuda.elementwise(
            'T z, T h_bar, T h, T gy', 'T g_z_x, T g_z_h, T g_h_x, T g_h, T g_hh', '''
            g_h = (1 - z) * gy;
            g_hh = z * (1 - h_bar * h_bar) * gy;
            g_h_x = g_hh;
            g_z_x = g_h * z * (h_bar - h);
            g_z_h = g_z_x;
            ''', 'compute_output_gru_bwd')(
            self.z, self.h_bar, h, gy[0])
        return g_z_x, g_z_h, g_h_x, g_h, g_hh,


def compute_output_GRU(z_x, z_h, h_x, h, hh):
    """ compute_output_GRU
    """
    return ComputeOutputGRU()(z_x, z_h, h_x, h, hh)


class ComputeOutputGRU2(function.Function):

    """SigmoidAPLusBByH function."""

#     def __init__(self):
#         self.use_cudnn = use_cudnn

    def check_type_forward(self, in_types):
        type_check.expect(in_types.size() == 3)
        type_check.expect(in_types[0].dtype == numpy.float32)
        type_check.expect(in_types[1].dtype == numpy.float32)
        type_check.expect(in_types[2].dtype == numpy.float32)

    def forward_cpu(self, x):
        z, h, hh = x
        self.h_bar = numpy.tanh(hh)
        h_new = (1 - z) * h + z * self.h_bar
        return h_new,

    def forward_gpu(self, x):
        z, h, hh = x
        self.h_bar, h_new = cuda.elementwise(
            'T z, T h, T hh',
            'T h_bar, T h_new',
            '''
                h_bar = tanh(hh);
                h_new = (1 - z) * h + z * h_bar;
                ''',
            'compute_output_gru_fwd')(z, h, hh)
        return h_new,

    def backward_cpu(self, x, gy):
        z, h, hh = x
        g_h = (1 - z) * gy[0]
        g_z = (self.h_bar - h) * gy[0]
        g_hh = z * (1 - self.h_bar * self.h_bar) * gy[0]
        return g_z, g_h, g_hh

    def backward_gpu(self, x, gy):
        z, h, hh = x
        g_z, g_h, g_hh = cuda.elementwise(
            'T z, T h_bar, T h, T gy', 'T g_z, T g_h, T g_hh',
            '''
            g_h = (1 - z) * gy;
            g_hh = z * (1 - h_bar * h_bar) * gy;
            g_z = g_h * (h_bar - h);
            ''',
            'compute_output_gru_bwd')(z, self.h_bar, h, gy[0])
        return g_z, g_h, g_hh,


def compute_output_GRU2(z, h, hh):
    """ compute_output_GRU
    """
    return ComputeOutputGRU2()(z, h, hh)


def sigm_a_plus_b_by_h(r_x, r_h, h):
    return h * (sigmoid.sigmoid(r_x + r_h))


def compute_output(z_x, z_h, h_x, h, hh):
    z = sigmoid.sigmoid(z_x + z_h)
    h_bar = tanh.tanh(h_x + hh)
    h_new = (1 - z) * h + z * h_bar
    return h_new


class GRUBase(link.Chain):

    def __init__(self, n_units, n_inputs=None, init=None, bias_init=None):
        if n_inputs is None:
            n_inputs = n_units
        super(GRUBase, self).__init__(
            W_r_z_h=linear.Linear(n_inputs, n_units * 3, initialW=init, initial_bias=bias_init),
            U_r_z=linear.Linear(n_units, n_units * 2, initialW=init, initial_bias=bias_init),
            #             W_r=linear.Linear(n_inputs, n_units),
            #             U_r = linear.Linear(n_units, n_units),
            #             W_z=linear.Linear(n_inputs, n_units),
            #             U_z=linear.Linear(n_units, n_units),
            #             W=linear.Linear(n_inputs, n_units),
            U=linear.Linear(n_units, n_units),
        )
        self.n_units = n_units

    def initialize_with_classic_implementation(self, gru):
        assert isinstance(gru, chainer.links.GRU)
        self.W_r_z_h.W.data[:self.n_units] = gru.W_r.W.data
        self.W_r_z_h.W.data[self.n_units: 2 * self.n_units] = gru.W_z.W.data
        self.W_r_z_h.W.data[2 * self.n_units:] = gru.W.W.data
        self.W_r_z_h.b.data[:self.n_units] = gru.W_r.b.data
        self.W_r_z_h.b.data[self.n_units: 2 * self.n_units] = gru.W_z.b.data
        self.W_r_z_h.b.data[2 * self.n_units:] = gru.W.b.data

        self.U_r_z.W.data[:self.n_units] = gru.U_r.W.data
        self.U_r_z.W.data[self.n_units:] = gru.U_z.W.data
        self.U_r_z.b.data[:self.n_units] = gru.U_r.b.data
        self.U_r_z.b.data[self.n_units:] = gru.U_z.b.data

        self.U.W.data[...] = gru.U.W.data
        self.U.b.data[...] = gru.U.b.data


class GRUBase2(link.Chain):
    def __init__(self, n_units, n_inputs=None):
        if n_inputs is None:
            n_inputs = n_units
        super(GRUBase2, self).__init__(
            W_r_z=linear.Linear(n_inputs + n_units, n_units * 2),
            W_h=linear.Linear(n_inputs + n_units, n_units),
        )
        self.n_units = n_units


import chainer.functions as F


def compute_GRU_out_2(z, h, hh):
    h_bar = F.tanh(hh)
    h_new = (1 - z) * h + z * h_bar
    return h_new


class GRU2(GRUBase2):

    """Stateless Gated Recurrent Unit function (GRU).
       """

    def faster_call(self, h, x):
        h_x = F.concat((h, x), axis=1)
        z_r = F.sigmoid(self.W_r_z(h_x))

        z, r = F.split_axis(z_r, (self.n_units,), axis=1)

        hr_x = F.concat((h * r, x), axis=1)
        h_bar = F.tanh(self.W_h(hr_x))

        h_new = (1 - z) * h + z * h_bar
        return h_new

    def faster_call2(self, h, x):
        h_x = F.concat((h, x), axis=1)
        z_r = F.sigmoid(self.W_r_z(h_x))

        z, r = F.split_axis(z_r, (self.n_units,), axis=1)

        hr_x = F.concat((h * r, x), axis=1)

        h_new = compute_GRU_out_2(z, h, self.W_h(hr_x))
        return h_new

    def faster_call3(self, h, x):
        h_x = F.concat((h, x), axis=1)
        z_r = F.sigmoid(self.W_r_z(h_x))

        z, r = F.split_axis(z_r, (self.n_units,), axis=1)

        hr_x = F.concat((h * r, x), axis=1)

        h_new = compute_output_GRU2(z, h, self.W_h(hr_x))
        return h_new


class GRU(GRUBase):

    """Stateless Gated Recurrent Unit function (GRU).
       """

    def classic_call(self, h, x):
        r = sigmoid.sigmoid(self.W_r(x) + self.U_r(h))
        z = sigmoid.sigmoid(self.W_z(x) + self.U_z(h))
        h_bar = tanh.tanh(self.W(x) + self.U(r * h))
        h_new = (1 - z) * h + z * h_bar
        return h_new

    def faster_call(self, h, x):
        r_z_h_x = self.W_r_z_h(x)
        r_x, z_x, h_x = split_axis(r_z_h_x, (self.n_units, self.n_units * 2), axis=1)
        assert r_x.data.shape[1] == self.n_units
        assert z_x.data.shape[1] == self.n_units
        assert h_x.data.shape[1] == self.n_units

        r_z_h = self.U_r_z(h)
        r_h, z_h = split_axis(r_z_h, (self.n_units,), axis=1)

        r = sigmoid.sigmoid(r_x + r_h)
        z = sigmoid.sigmoid(z_x + z_h)
        h_bar = tanh.tanh(h_x + self.U(r * h))
        h_new = (1 - z) * h + z * h_bar
        return h_new

    def faster_call2(self, h, x):
        r_z_h_x = self.W_r_z_h(x)

        r_z_h = self.U_r_z(h)

        r_x, z_x, h_x = split_axis(r_z_h_x, (self.n_units, self.n_units * 2), axis=1)
        assert r_x.data.shape[1] == self.n_units
        assert z_x.data.shape[1] == self.n_units
        assert h_x.data.shape[1] == self.n_units

        r_h, z_h = split_axis(r_z_h, (self.n_units,), axis=1)
#         r = sigmoid.sigmoid(r_x + r_h)
#         z = sigmoid.sigmoid(z_x + z_h)
#         h_bar = tanh.tanh(h_x + self.U(sigm_a_plus_b_by_h(r_x, r_h, h)))
#         h_new = (1 - z) * h + z * h_bar
#         return h_new

        return compute_output_GRU(z_x, z_h, h_x, h, self.U(sigm_a_plus_b_by_h_fast(r_x, r_h, h)))

    def faster_call3(self, h, x):
        device = chainer.cuda.get_device(h.data)
        with device:
            stream1 = chainer.cuda.Stream(1)
            stream2 = chainer.cuda.Stream(2)

            chainer.cuda.cublas.setStream(device.cublas_handle, stream1.ptr)
            r_z_h_x = self.W_r_z_h(x)

            chainer.cuda.cublas.setStream(device.cublas_handle, stream2.ptr)
            r_z_h = self.U_r_z(h)

            stream1.synchronize()
            stream2.synchronize()

            r_x, z_x, h_x = split_axis(r_z_h_x, (self.n_units, self.n_units * 2), axis=1)
            assert r_x.data.shape[1] == self.n_units
            assert z_x.data.shape[1] == self.n_units
            assert h_x.data.shape[1] == self.n_units

            r_h, z_h = split_axis(r_z_h, (self.n_units,), axis=1)
    #         r = sigmoid.sigmoid(r_x + r_h)
    #         z = sigmoid.sigmoid(z_x + z_h)
    #         h_bar = tanh.tanh(h_x + self.U(sigm_a_plus_b_by_h(r_x, r_h, h)))
    #         h_new = (1 - z) * h + z * h_bar
    #         return h_new

            return compute_output_GRU(z_x, z_h, h_x, h, self.U(sigm_a_plus_b_by_h_fast(r_x, r_h, h)))

    __call__ = faster_call2

    def initialize_with_classic_implementation(self, gru):
        """ Initialize with the parameters of a non optimized GRU. Useful for testing / model transfer. """
        assert isinstance(gru, chainer.links.connection.GRU)
        self.W_r_z_h.W.data[:self.n_units] = gru.W_r.W.data
        self.W_r_z_h.W.data[self.n_units: 2 * self.n_units] = gru.W_z.W.data
        self.W_r_z_h.W.data[2 * self.n_units:] = gru.W.W.data
        self.W_r_z_h.b.data[:self.n_units] = gru.W_r.b.data
        self.W_r_z_h.b.data[self.n_units: 2 * self.n_units] = gru.W_z.b.data
        self.W_r_z_h.b.data[2 * self.n_units:] = gru.W.b.data

        self.U_r_z.W.data[:self.n_units] = gru.U_r.W.data
        self.U_r_z.W.data[self.n_units:] = gru.U_z.W.data
        self.U_r_z.b.data[:self.n_units] = gru.U_r.b.data
        self.U_r_z.b.data[self.n_units:] = gru.U_z.b.data

        self.U.W.data[...] = gru.U.W.data
        self.U.b.data[...] = gru.U.b.data


import numpy as np
from chainer import Variable, cuda
from chainer import gradient_check
# import chainer.function_hooks
import operator


def commandline():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--gpu")
    parser.add_argument("--fast", type=int)
    parser.add_argument("--fast2", type=int)
    parser.add_argument("--test_perf", default=False, action="store_true")
    parser.add_argument("--test_correctness", default=False, action="store_true")
    parser.add_argument("--mb_size", type=int, default=64)
    parser.add_argument("--h_size", type=int, default=512)
    parser.add_argument("--i_size", type=int, default=512)
    args = parser.parse_args()

    if args.test_perf:
        test_perf(args)

    if args.test_correctness:
        test_correctness(args)


def test_correctness(args):
    ni = 100
    no = 100
    mb_size = 8

    gru_classic = chainer.links.GRU(ni, no)
    my_gru = GRU(ni, no)
    my_gru.initialize_with_classic_implementation(gru_classic)

    x = np.random.randn(mb_size, ni).astype(np.float32)
    st = np.random.randn(mb_size, no).astype(np.float32)

    grad_output = np.random.randn(mb_size, no).astype(np.float32)

    if args.gpu is not None:
        my_gru = my_gru.to_gpu(args.gpu)
        gru_classic = gru_classic.to_gpu(args.gpu)
        x = cuda.to_gpu(x, args.gpu)
        st = cuda.to_gpu(st, args.gpu)
        grad_output = cuda.to_gpu(grad_output, args.gpu)

    x_v = Variable(x)
    x_v2 = Variable(x)

    st_v = Variable(st)
    st_v2 = Variable(st)

    with chainer.function_hooks.PrintHook():
        with chainer.function_hooks.TimerHook() as m:
            st_classic = gru_classic(st_v, x_v)
            print m
            print(m.total_time())
            print sorted(m.call_history, key=operator.itemgetter(1))

        with chainer.function_hooks.TimerHook() as m:
            my_st = my_gru.faster_call2(st_v2, x_v2)
            print m
            print(m.total_time())
            print sorted(m.call_history, key=operator.itemgetter(1))

    print np.max(np.abs(cuda.to_cpu(my_st.data - st_classic.data)))
    gradient_check.assert_allclose(my_st.data, st_classic.data)

    my_st.grad = grad_output
    st_classic.grad = grad_output

    my_gru.zerograds()
    gru_classic.zerograds()

    my_st.backward()
    st_classic.backward()

    gradient_check.assert_allclose(x_v.grad, x_v2.grad)
    gradient_check.assert_allclose(st_v.grad, st_v2.grad)

    gradient_check.assert_allclose(gru_classic.U.W.grad, my_gru.U.W.grad)
    gradient_check.assert_allclose(gru_classic.U.b.grad, my_gru.U.b.grad)


def test_perf(args):
    ni = args.i_size
    no = args.h_size
    mb_size = args.mb_size

    x = np.random.randn(mb_size, ni).astype(np.float32)
    st = np.random.randn(mb_size, no).astype(np.float32)

    if args.fast2 is not None:
        gru_model = GRU2(no, ni)
        if args.gpu is not None:
            gru_model = gru_model.to_gpu(args.gpu)

        if args.fast2 == 1:
            gru = gru_model.faster_call
        elif args.fast2 == 2:
            gru = gru_model.faster_call2
        elif args.fast2 == 3:
            gru = gru_model.faster_call3

    elif args.fast is not None:
        gru_model = GRU(no, ni)
        if args.gpu is not None:
            gru_model = gru_model.to_gpu(args.gpu)
        if args.fast == 0:
            gru = gru_model.classic_call
        elif args.fast == 1:
            gru = gru_model.faster_call
        elif args.fast == 2:
            gru = gru_model.faster_call2
        elif args.fast == 3:
            gru = gru_model.faster_call3
    else:
        gru = chainer.links.GRU(no, ni)
        if args.gpu is not None:
            gru = gru.to_gpu(args.gpu)
#     gru = GRU(no, ni)

    if args.gpu is not None:
        x = cuda.to_gpu(x, args.gpu)
        st = cuda.to_gpu(st, args.gpu)

    print "start"
    for _ in xrange(300):
        x_v = Variable(x)
        st_v = Variable(st)
        for _ in xrange(20):
            st_v = gru(st_v, x_v)
        loss = chainer.functions.sum(st_v)
        loss.backward()
    print "done"


def test2():
    import argparse
    parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)
    parser.add_argument("--gpu")
    args = parser.parse_args()

    ni = 10
    no = 10
    mb_size = 32
    import numpy as np
    from chainer import Variable, cuda

    x = np.random.randn(mb_size, ni).astype(np.float32)
    st = np.random.randn(mb_size, no).astype(np.float32)

    gru_model = GRU(no, ni)
#     gru = GRU(no, ni)

    if args.gpu is not None:
        x = cuda.to_gpu(x, args.gpu)
        st = cuda.to_gpu(st, args.gpu)
        gru_model = gru_model.to_gpu(args.gpu)

    x_v = Variable(x)
    st_v = Variable(st)

    gru_model.zerograds()
#     r1 = gru_model.classic_call(st_v, x_v)
    r2 = gru_model.faster_call(st_v, x_v)
    r3 = gru_model.faster_call2(st_v, x_v)

    print r2.data - r2.data
    print r3.data - r3.data

    from chainer import gradient_check

    r3 = r2
    r3.grad = np.random.randn(mb_size, no).astype(np.float32)
    r3.backward()

    def f():
        return (gru_model.faster_call(st_v, x_v).data,)

    g_st, g_x = gradient_check.numerical_grad(f, (st_v.data, x_v.data), (r3.grad,))
    print np.max(np.abs(g_st - st_v.grad))
    print np.max(np.abs(g_x - x_v.grad))
    gradient_check.assert_allclose(g_st, st_v.grad, atol=1e-3)
    gradient_check.assert_allclose(g_x, x_v.grad, atol=1e-3)


if __name__ == '__main__':
    commandline()



#!/usr/bin/env python
"""train.py: Train a RNNSearch Model"""
__author__ = "Fabien Cromieres"
__license__ = "undecided"
__version__ = "1.0"
__email__ = "fabien.cromieres@gmail.com"
__status__ = "Development"

import chainer
from chainer import cuda, optimizers, serializers

from training import train_on_data
from nmt_chainer.dataprocessing.indexer import Indexer

import logging
import json
import os.path
import gzip
import sys
import pprint
# import h5py

from nmt_chainer.utilities.utils import ensure_path
# , make_batch_src_tgt, make_batch_src, minibatch_provider, compute_bleu_with_unk_as_wrong,de_batch
# from evaluation import (
#                   greedy_batch_translate, convert_idx_to_string,
#                   compute_loss_all, translate_to_file, sample_once)

import nmt_chainer.models.attention
import nmt_chainer.models.encoder_decoder

import nmt_chainer.models.rnn_cells as rnn_cells
import nmt_chainer.dataprocessing.processors as processors

import nmt_chainer.utilities.profiling_tools as profiling_tools

logging.basicConfig()
log = logging.getLogger("rnns:train")
log.setLevel(logging.INFO)


def generate_lexical_probability_dictionary_indexed(lexical_probability_dictionary_all, src_indexer, tgt_indexer):
    log.info("computing lexical_probability_dictionary_indexed")
    lexical_probability_dictionary_indexed = {}
    for ws in lexical_probability_dictionary_all:
        ws_idx = src_indexer.convert([ws])[0]
        if ws_idx in lexical_probability_dictionary_indexed:
            assert src_indexer.is_unk_idx(ws_idx)
        else:
            lexical_probability_dictionary_indexed[ws_idx] = {}
        for wt in lexical_probability_dictionary_all[ws]:
            wt_idx = tgt_indexer.convert([wt])[0]
            if wt_idx in lexical_probability_dictionary_indexed[ws_idx]:
                assert src_indexer.is_unk_idx(
                    ws_idx) or tgt_indexer.is_unk_idx(wt_idx)
                lexical_probability_dictionary_indexed[ws_idx][wt_idx] += lexical_probability_dictionary_all[ws][wt]
            else:
                lexical_probability_dictionary_indexed[ws_idx][wt_idx] = lexical_probability_dictionary_all[ws][wt]
    return lexical_probability_dictionary_indexed


def create_encdec_from_config_dict(config_dict, src_indexer, tgt_indexer):
    Ei = config_dict["Ei"]
    Hi = config_dict["Hi"]
    Eo = config_dict["Eo"]
    Ho = config_dict["Ho"]
    Ha = config_dict["Ha"]
    Hl = config_dict["Hl"]

    Vi = len(src_indexer)  # + UNK
    Vo = len(tgt_indexer)  # + UNK

    encoder_cell_type = config_dict.get("encoder_cell_type", "gru")
    decoder_cell_type = config_dict.get("decoder_cell_type", "gru")

    use_bn_length = config_dict.get("use_bn_length", None)

    # Selecting Attention type
    attn_cls = nmt_chainer.models.attention.AttentionModule
    if config_dict.get("use_accumulated_attn", False):
        raise NotImplemented
    if config_dict.get("use_deep_attn", False):
        attn_cls = nmt_chainer.models.attention.DeepAttentionModule

    init_orth = config_dict.get("init_orth", False)

    if "lexical_probability_dictionary" in config_dict and config_dict["lexical_probability_dictionary"] is not None:
        log.info("opening lexical_probability_dictionary %s" % config_dict["lexical_probability_dictionary"])
        lexical_probability_dictionary_all = json.load(gzip.open(config_dict["lexical_probability_dictionary"], "rb"))

        lexical_probability_dictionary = generate_lexical_probability_dictionary_indexed(
            lexical_probability_dictionary_all, src_indexer, tgt_indexer)
    else:
        lexical_probability_dictionary = None
    lex_epsilon = config_dict.get("lexicon_prob_epsilon", 0.001)

    # Creating encoder/decoder
    encdec = nmt_chainer.models.encoder_decoder.EncoderDecoder(Vi, Ei, Hi, Vo + 1, Eo, Ho, Ha, Hl, use_bn_length=use_bn_length,
                                                               attn_cls=attn_cls,
                                                               init_orth=init_orth,
                                                               encoder_cell_type=rnn_cells.create_cell_model_from_config(encoder_cell_type),
                                                               decoder_cell_type=rnn_cells.create_cell_model_from_config(decoder_cell_type),
                                                               lexical_probability_dictionary=lexical_probability_dictionary,
                                                               lex_epsilon=lex_epsilon)

    return encdec


def create_encdec_and_indexers_from_config_dict(config_dict, src_indexer=None, tgt_indexer=None, load_config_model="no"):
    assert load_config_model in "yes no if_exists".split()

    if src_indexer is None or tgt_indexer is None:
        voc_fn = config_dict.data["voc"]
        log.info("loading voc from %s" % voc_fn)
#         src_voc, tgt_voc = json.load(open(voc_fn))

        bi_idx = processors.load_pp_pair_from_file(voc_fn)

    if src_indexer is None:
        src_indexer = bi_idx.src_processor()

    if tgt_indexer is None:
        tgt_indexer = bi_idx.tgt_processor()

#     tgt_voc = None
#     src_voc = None

    encdec = create_encdec_from_config_dict(config_dict["model"], src_indexer, tgt_indexer)

    eos_idx = len(tgt_indexer)

    if load_config_model != "no":
        if "model_parameters" not in config_dict:
            if load_config_model == "yes":
                log.error("cannot find model parameters in config file")
                raise ValueError(
                    "Config file do not contain model_parameters section")
        else:
            if config_dict.model_parameters.type == "model":
                model_filename = config_dict.model_parameters.filename
                log.info(
                    "loading model parameters from file specified by config file:%s" %
                    model_filename)
                serializers.load_npz(model_filename, encdec)
            else:
                if load_config_model == "yes":
                    log.error(
                        "model parameters in config file is of type snapshot, not model")
                    raise ValueError("Config file model is not of type model")

    return encdec, eos_idx, src_indexer, tgt_indexer


def load_voc_and_update_training_config(config_training):
    data_prefix = config_training["training_management"]["data_prefix"]
    voc_fn = data_prefix + ".voc"
    data_fn = data_prefix + ".data.json.gz"

    log.info("loading voc from %s" % voc_fn)
#     src_voc, tgt_voc = json.load(open(voc_fn))

    bi_idx = processors.load_pp_pair_from_file(voc_fn)

    src_indexer, tgt_indexer = bi_idx.src_processor(), bi_idx.tgt_processor()
#     src_indexer = processors.PreProcessor.make_from_serializable(src_voc)
#     tgt_indexer = processors.PreProcessor.make_from_serializable(tgt_voc)
#     tgt_voc = None
#     src_voc = None

#     Vi = len(src_voc) + 1 # + UNK
#     Vo = len(tgt_voc) + 1 # + UNK

    Vi = len(src_indexer)  # + UNK
    Vo = len(tgt_indexer)  # + UNK

    config_training.add_section("data", keep_at_bottom="metadata", overwrite=False)
    config_training["data"]["data_fn"] = data_fn
    config_training["data"]["Vi"] = Vi
    config_training["data"]["Vo"] = Vo
    config_training["data"]["voc"] = voc_fn

    config_training.set_readonly()

    return src_indexer, tgt_indexer


def do_train(config_training):

    src_indexer, tgt_indexer = load_voc_and_update_training_config(config_training)

    save_prefix = config_training.training_management.save_prefix

    output_files_dict = {}
    output_files_dict["train_config"] = save_prefix + ".train.config"
    output_files_dict["model_ckpt"] = save_prefix + ".model." + "ckpt" + ".npz"
    output_files_dict["model_final"] = save_prefix + \
        ".model." + "final" + ".npz"
    output_files_dict["model_best"] = save_prefix + ".model." + "best" + ".npz"
    output_files_dict["model_best_loss"] = save_prefix + ".model." + "best_loss" + ".npz"

#     output_files_dict["model_ckpt_config"] = save_prefix + ".model." + "ckpt" + ".config"
#     output_files_dict["model_final_config"] = save_prefix + ".model." + "final" + ".config"
#     output_files_dict["model_best_config"] = save_prefix + ".model." + "best" + ".config"
#     output_files_dict["model_best_loss_config"] = save_prefix + ".model." + "best_loss" + ".config"

    output_files_dict["test_translation_output"] = save_prefix + ".test.out"
    output_files_dict["test_src_output"] = save_prefix + ".test.src.out"
    output_files_dict["dev_translation_output"] = save_prefix + ".dev.out"
    output_files_dict["dev_src_output"] = save_prefix + ".dev.src.out"
    output_files_dict["valid_translation_output"] = save_prefix + ".valid.out"
    output_files_dict["valid_src_output"] = save_prefix + ".valid.src.out"
    output_files_dict["sqlite_db"] = save_prefix + ".result.sqlite"
    output_files_dict["optimizer_ckpt"] = save_prefix + ".optimizer." + "ckpt" + ".npz"
    output_files_dict["optimizer_final"] = save_prefix + ".optimizer." + "final" + ".npz"

    save_prefix_dir, save_prefix_fn = os.path.split(save_prefix)
    ensure_path(save_prefix_dir)

    already_existing_files = []
    for key_info, filename in output_files_dict.iteritems():  # , valid_data_fn]:
        if os.path.exists(filename):
            already_existing_files.append(filename)
    if len(already_existing_files) > 0:
        print "Warning: existing files are going to be replaced / updated: ", already_existing_files
        if not config_training.training_management.force_overwrite:
            raw_input("Press Enter to Continue")

    save_train_config_fn = output_files_dict["train_config"]
    log.info("Saving training config to %s" % save_train_config_fn)
    config_training.save_to(save_train_config_fn)
#     json.dump(config_training, open(save_train_config_fn, "w"), indent=2, separators=(',', ': '))

    Vi = len(src_indexer)  # + UNK
    Vo = len(tgt_indexer)  # + UNK

    eos_idx = Vo

    data_fn = config_training.data.data_fn

    log.info("loading training data from %s" % data_fn)
    training_data_all = json.load(gzip.open(data_fn, "rb"))

    training_data = training_data_all["train"]

    log.info("loaded %i sentences as training data" % len(training_data))

    if "test" in training_data_all:
        test_data = training_data_all["test"]
        log.info("Found test data: %i sentences" % len(test_data))
    else:
        test_data = None
        log.info("No test data found")

    if "dev" in training_data_all:
        dev_data = training_data_all["dev"]
        log.info("Found dev data: %i sentences" % len(dev_data))
    else:
        dev_data = None
        log.info("No dev data found")

    if "valid" in training_data_all:
        valid_data = training_data_all["valid"]
        log.info("Found valid data: %i sentences" % len(valid_data))
    else:
        valid_data = None
        log.info("No valid data found")

    max_src_tgt_length = config_training.training_management.max_src_tgt_length
    if max_src_tgt_length is not None:
        log.info("filtering sentences of length larger than %i" % (max_src_tgt_length))
        filtered_training_data = []
        nb_filtered = 0
        for src, tgt in training_data:
            if len(src) <= max_src_tgt_length and len(
                    tgt) <= max_src_tgt_length:
                filtered_training_data.append((src, tgt))
            else:
                nb_filtered += 1
        log.info("filtered %i sentences of length larger than %i" % (nb_filtered, max_src_tgt_length))
        training_data = filtered_training_data

    if not config_training.training.no_shuffle_of_training_data:
        log.info("shuffling")
        import random
        random.shuffle(training_data)
        log.info("done")

    encdec, _, _, _ = create_encdec_and_indexers_from_config_dict(config_training,
                                                                  src_indexer=src_indexer, tgt_indexer=tgt_indexer,
                                                                  load_config_model="if_exists" if config_training.training_management.resume else "no")
#     create_encdec_from_config_dict(config_training.model, src_indexer, tgt_indexer,
#                             load_config_model = "if_exists" if config_training.training_management.resume else "no")

#     if config_training.training_management.resume:
#         if "model_parameters" not in config_training:
#             log.error("cannot find model parameters in config file")
#         if config_training.model_parameters.type == "model":
#             model_filename = config_training.model_parameters.filename
#             log.info("resuming from model parameters %s" % model_filename)
#             serializers.load_npz(model_filename, encdec)

    if config_training.training_management.load_model is not None:
        log.info("loading model parameters from %s", config_training.training_management.load_model)
        serializers.load_npz(config_training.training_management.load_model, encdec)

    gpu = config_training.training_management.gpu
    if gpu is not None:
        encdec = encdec.to_gpu(gpu)

    if config_training.training.optimizer == "adadelta":
        optimizer = optimizers.AdaDelta()
    elif config_training.training.optimizer == "adam":
        optimizer = optimizers.Adam()
    elif config_training.training.optimizer == "adagrad":
        optimizer = optimizers.AdaGrad(lr=config_training.training.learning_rate)
    elif config_training.training.optimizer == "sgd":
        optimizer = optimizers.SGD(lr=config_training.training.learning_rate)
    elif config_training.training.optimizer == "momentum":
        optimizer = optimizers.MomentumSGD(lr=config_training.training.learning_rate,
                                           momentum=config_training.training.momentum)
    elif config_training.training.optimizer == "nesterov":
        optimizer = optimizers.NesterovAG(lr=config_training.training.learning_rate,
                                          momentum=config_training.training.momentum)
    elif config_training.training.optimizer == "rmsprop":
        optimizer = optimizers.RMSprop(lr=config_training.training.learning_rate)
    elif config_training.training.optimizer == "rmspropgraves":
        optimizer = optimizers.RMSpropGraves(lr=config_training.training.learning_rate,
                                             momentum=config_training.training.momentum)
    else:
        raise NotImplemented

    with cuda.get_device(gpu):
        optimizer.setup(encdec)

    if config_training.training.l2_gradient_clipping is not None and config_training.training.l2_gradient_clipping > 0:
        optimizer.add_hook(chainer.optimizer.GradientClipping(
            config_training.training.l2_gradient_clipping))

    if config_training.training.hard_gradient_clipping is not None and config_training.training.hard_gradient_clipping > 0:
        optimizer.add_hook(chainer.optimizer.GradientHardClipping(
            *config_training.training.hard_gradient_clipping))

    if config_training.training.weight_decay is not None:
        optimizer.add_hook(
            chainer.optimizer.WeightDecay(
                config_training.training.weight_decay))

    if config_training.training_management.load_optimizer_state is not None:
        with cuda.get_device(gpu):
            log.info("loading optimizer parameters from %s", config_training.training_management.load_optimizer_state)
            serializers.load_npz(config_training.training_management.load_optimizer_state, optimizer)

    if config_training.training_management.timer_hook:
        timer_hook = profiling_tools.MyTimerHook
    else:
        import contextlib

        @contextlib.contextmanager
        def timer_hook():
            yield

    import training_chainer
    with cuda.get_device(gpu):
        with timer_hook() as timer_infos:

            if config_training.training_management.max_nb_iters is not None:
                stop_trigger = (
                    config_training.training_management.max_nb_iters,
                    "iteration")
                if config_training.training_management.max_nb_epochs is not None:
                    log.warn(
                        "max_nb_iters and max_nb_epochs both specified. Only max_nb_iters will be considered.")
            elif config_training.training_management.max_nb_epochs is not None:
                stop_trigger = (
                    config_training.training_management.max_nb_epochs, "epoch")
            else:
                stop_trigger = None
            training_chainer.train_on_data_chainer(encdec, optimizer, training_data, output_files_dict,
                                                   src_indexer, tgt_indexer, eos_idx=eos_idx,
                                                   config_training=config_training,
                                                   stop_trigger=stop_trigger,
                                                   test_data=test_data, dev_data=dev_data, valid_data=valid_data
                                                   )


#
#     import sys
#     sys.exit(0)
#     with cuda.get_device(args.gpu):
# #         with MyTimerHook() as timer:
# #             try:
#                 train_on_data(encdec, optimizer, training_data, output_files_dict,
#                       src_indexer, tgt_indexer, eos_idx = eos_idx,
#                       mb_size = args.mb_size,
#                       nb_of_batch_to_sort = args.nb_batch_to_sort,
#                       test_data = test_data, dev_data = dev_data, valid_data = valid_data, gpu = args.gpu, report_every = args.report_every,
#                       randomized = args.randomized_data, reverse_src = args.reverse_src, reverse_tgt = args.reverse_tgt,
#                       max_nb_iters = args.max_nb_iters, do_not_save_data_for_resuming = args.no_resume,
#                       noise_on_prev_word = args.noise_on_prev_word, curiculum_training = args.curiculum_training,
#                       use_previous_prediction = args.use_previous_prediction, no_report_or_save = args.no_report_or_save,
#                       use_memory_optimization = args.use_memory_optimization,
#                       sample_every = args.sample_every,
#                       use_reinf = args.use_reinf,
#                       save_ckpt_every = args.save_ckpt_every
# #                     lexical_probability_dictionary = lexical_probability_dictionary,
# #                     V_tgt = Vo + 1,
# #                     lexicon_prob_epsilon = args.lexicon_prob_epsilon
#                       )
# #             finally:
# #                 print timer
# #                 timer.print_sorted()
# #                 print "total time:"
# #                 print(timer.total_time())
#
#



#!/usr/bin/python
# -*- coding: utf-8 -*-
# Author: Rico Sennrich

"""Use operations learned with learn_bpe.py to encode a new text.
The text will not be smaller, but use only a fixed vocabulary, with rare words
encoded as variable-length sequences of subword units.

Reference:
Rico Sennrich, Barry Haddow and Alexandra Birch (2015). Neural Machine Translation of Rare Words with Subword Units.
Proceedings of the 54th Annual Meeting of the Association for Computational Linguistics (ACL 2016). Berlin, Germany.
"""

# Dec. 2016: Edited by Raj Dabre

from __future__ import unicode_literals, division

import sys
import codecs
import argparse
from collections import defaultdict

# hack for python2/3 compatibility
from io import open
argparse.open = open

# python 2/3 compatibility
# if sys.version_info < (3, 0):
#   sys.stderr = codecs.getwriter('UTF-8')(sys.stderr)
#   sys.stdout = codecs.getwriter('UTF-8')(sys.stdout)
#   sys.stdin = codecs.getreader('UTF-8')(sys.stdin)

import codecs


class BPE(object):

    def __init__(self, codes, separator='__'):

        with codecs.open(codes.name, encoding='utf-8') as codes:
            self.bpe_codes = [tuple(item.split()) for item in codes]

        # some hacking to deal with duplicates (only consider first instance)
        self.bpe_codes = dict([(code, i) for (i, code) in reversed(list(enumerate(self.bpe_codes)))])

        self.separator = separator
        self.cache = {}

    def segment(self, sentence):
        """segment single sentence (whitespace-tokenized string) with BPE encoding"""

        output = []
        for word in sentence.split():
            new_word = encode(word, self.bpe_codes, self.cache)
            for item in new_word[:-1]:
                output.append(item + self.separator)
            output.append(new_word[-1])
        # print output
        return ' '.join(output)

    def segment_splitted(self, sentence):
        """segment single sentence (as a sequence of words) with BPE encoding"""

        output = []
        for word in sentence:
            new_word = encode(word, self.bpe_codes, self.cache)
            for item in new_word[:-1]:
                output.append(item + self.separator)
            output.append(new_word[-1])
        # print output
        return output


def create_parser():
    parser = argparse.ArgumentParser(
        formatter_class=argparse.RawDescriptionHelpFormatter,
        description="learn BPE-based word segmentation")

    parser.add_argument(
        '--input', '-i', type=argparse.FileType('r'), default=sys.stdin,
        metavar='PATH',
        help="Input file (default: standard input).")
    parser.add_argument(
        '--codes', '-c', type=argparse.FileType('r'), metavar='PATH',
        required=True,
        help="File with BPE codes (created by learn_bpe.py).")
    parser.add_argument(
        '--output', '-o', type=argparse.FileType('w'), default=sys.stdout,
        metavar='PATH',
        help="Output file (default: standard output)")
    parser.add_argument(
        '--separator',
        '-s',
        type=str,
        default='@@',
        metavar='STR',
        help="Separator between non-final subword units (default: '%(default)s'))")

    return parser


def get_pairs(word):
    """Return set of symbol pairs in a word.

    word is represented as tuple of symbols (symbols being variable-length strings)
    """
    pairs = set()
    prev_char = word[0]
    for char in word[1:]:
        pairs.add((prev_char, char))
        prev_char = char
    return pairs


def encode(orig, bpe_codes, cache):
    """Encode word based on list of BPE merge operations, which are applied consecutively
    """
    if orig in cache:
        return cache[orig]

    word = tuple(orig) + ('</w>',)
    pairs = get_pairs(word)

    while True:
        bigram = min(pairs, key=lambda pair: bpe_codes.get(pair, float('inf')))
        if bigram not in bpe_codes:
            break
        first, second = bigram
        new_word = []
        i = 0
        while i < len(word):
            try:
                j = word.index(first, i)
                new_word.extend(word[i:j])
                i = j
            except BaseException:
                new_word.extend(word[i:])
                break

            if word[i] == first and i < len(word) - 1 and word[i + 1] == second:
                new_word.append(first + second)
                i += 2
            else:
                new_word.append(word[i])
                i += 1
        new_word = tuple(new_word)
        word = new_word
        if len(word) == 1:
            break
        else:
            pairs = get_pairs(word)

    # don't print end-of-word symbols
    if word[-1] == '</w>':
        word = word[:-1]
    elif word[-1].endswith('</w>'):
        word = word[:-1] + (word[-1].replace('</w>', ''),)

    cache[orig] = word
    return word


if __name__ == '__main__':
    parser = create_parser()
    args = parser.parse_args()

    bpe = BPE(args.codes, args.separator)

    for line in args.input:
        args.output.write(bpe.segment(line).strip())
        args.output.write('\n')
